__author__ = "hugo.inzirillo"

from collections import OrderedDict
from dataclasses import dataclass
from time import time, sleep

import requests

ITERATION = 10

response = requests.get("https://www.bitmex.com/api/v1/orderBook/L2?symbol=eth&depth=100000").json()
response2 = requests.get("https://www.bitmex.com/api/v1/orderBook/L2?symbol=xbt&depth=100000").json()
response = response +response2

def time_this(original_function):
    def new_function(*args, **kwargs):
        import datetime
        before = datetime.datetime.now()
        x = original_function(*args, **kwargs)
        after = datetime.datetime.now()
        print("Elapsed Time = {0}".format(after - before))
        return x

    return new_function


class Quote(object):
    def __init__(self, product_code, ts, close):
        self.__product_code = product_code
        self.__ts = ts
        self.__close = close


"https://www.bitmex.com/api/v1/orderBook/L2?symbol=eth&depth=100000"


@dataclass(init=False)
class BitMexOrder:
    symbol: str
    id: int
    side: str
    price: float

    def from_dict(self, dict_object: dict):
        if isinstance(dict_object, dict):
            self.__dict__ = dict_object
            return self

    def __hash__(self):
        return hash((self.symbol, self.id, self.side, self.price))


class QuoteSlottedPk(object):
    __slots__ = ("product_code", "ts")

    def __init__(self, product_code: str, ts: float):
        self.product_code = product_code
        self.ts = ts

    def __hash__(self):
        return hash((self.product_code, self.ts))

    def __eq__(self, other):
        return self.product_code == other.product_code and self.ts == other.ts if isinstance(other,
                                                                                             self.__class__) else False


class QuoteSlotted(object):
    __slots__ = ("close", "pk")

    def __init__(self, pk, close):
        self.close = close
        self.pk = pk

    def __hash__(self):
        return hash((self.pk, self.close))

    def __eq__(self, other):
        return self.pk.__eq__(other.pk) if isinstance(other, self.__class__) else False


def test3():
    res = OrderedDict()
    for iter in range(ITERATION):
        quote = QuoteSlotted(QuoteSlottedPk("Okayebifibzeifbziefbizuebfizebifbziuef", time()), iter)
        res[quote.pk] = quote
        sleep(0.01)
    return res


@time_this
def bitmex2(response):
    return (BitMexOrder().from_dict(item) for item in response)


if __name__ == '__main__':
    gen = test3()

    carnet2 = bitmex2(response)





__author__ = "hugo.inzirillo"

from dataclasses import dataclass
from typing import List


@dataclass
class SignalComputation:
    """
    Position Service
    Mapping of Java Objects
    """
    productCode: List[str] = None
    underlyingCode: List[str] = None
    ts: str = None
    value: float = None

    def from_dict(self, dict_object: dict):
        if isinstance(dict_object, dict):
            self.__dict__ = dict_object
            return self



#!/usr/bin/env python3
# coding: utf-8

import ccxt
import urllib.parse,  urllib.request
from urllib.parse import urljoin, urlencode
import json, hashlib, hmac, time
from datetime import datetime
import requests
import calendar

import pandas as pd

from datetime import  timedelta

class NaPoleonBinanceFutureBot(object):

    def __init__(self, BINANCE_PUBLIC, BINANCE_PRIVATE):
        self.apiKey = BINANCE_PUBLIC
        self.secret = BINANCE_PRIVATE
        Binancecle2 = {
            'api_key': BINANCE_PUBLIC,
            'api_secret': BINANCE_PRIVATE,
        }
        print('Connexion to Binance Future')
        exchange_id = 'binance'
        exchange_class = getattr(ccxt, exchange_id)
        self.exchange = exchange_class({
            'apiKey': Binancecle2['api_key'],
            'secret': Binancecle2['api_secret'],
            'timeout': 10000,
            'enableRateLimit': True,
            'options': {'defaultType': 'future'},
        })

    def cancel_all_order(self, pair):
        Cleaner = self.exchange.fetch_open_orders(pair)
        L = len(Cleaner)
        if L > 0:
            for i in range(0, L):
                self.exchange.cancel_order(Cleaner[i]['info']['orderId'], 'BTC/USDT')

    def cancel_all_orders(self):
        self.cancel_all_order('BTC/USDT')
        self.cancel_all_order('ETH/USDT')
        self.cancel_all_order('LTC/USDT')
        self.cancel_all_order('EOS/USDT')
        self.cancel_all_order('XRP/USDT')
        self.cancel_all_order('BCH/USDT')
        self.cancel_all_order('TRX/USDT')


    def get_futures_balance_history(self, start_time, end_time):
        startTime = int(start_time.timestamp() * 1000)
        endTime = int(end_time.timestamp() * 1000)

        BASE_URL = 'https://api.binance.com'
        PATH = '/sapi/v1/accountSnapshot'

        headers = {
            'X-MBX-APIKEY': self.apiKey
        }
        timestamp = int(time.time() * 1000)

        params = {
            'type': 'FUTURES',
            'recvWindow': 60000,
            'timestamp': timestamp,
            'startTime': startTime,
            'endTime': endTime,
            'limit':30

        }
        query_string = urllib.parse.urlencode(params)
        params['signature'] = hmac.new(self.secret.encode('utf-8'), query_string.encode('utf-8'),
                                       hashlib.sha256).hexdigest()
        url = urljoin(BASE_URL, PATH)
        print(url)
        r = requests.get(url, headers=headers, params=params)
        dataSet = r.json()

        snapshots = dataSet['snapshotVos']

        concatenated_df = None
        for me_snapshot in snapshots:
            data_df = pd.DataFrame(me_snapshot['data']['assets'])
            data_df['updateTime'] = me_snapshot['updateTime']
            if concatenated_df is None:
                concatenated_df = data_df.copy()
            else :
                concatenated_df = pd.concat([concatenated_df, data_df])

        concatenated_df['date'] = pd.to_datetime(concatenated_df['updateTime'], unit='ms')
        concatenated_df['date'] = concatenated_df['date'] + timedelta(seconds=1)

        return concatenated_df



    def get_futures_balance_snapshot(self, request_date):
        request_date = request_date.replace(hour=0, minute=0, second=0, microsecond=0)
        start_time = request_date
        end_time = request_date + timedelta(days = 1)
        result_df = self.get_futures_balance_history(start_time, end_time)
        result_df = result_df[result_df['date'] == request_date]
        return result_df

    def get_futures_positions_margin_change_history(self, assets, start_time, end_time):
        positions_margin_change_df = None
        for asset in assets:
            if positions_margin_change_df is None:
                positions_margin_change_df = self.get_futures_position_margin_change_history(asset, start_time, end_time)
            else:
                add_transfer_df = self.get_futures_position_margin_change_history(asset, start_time, end_time)
                if add_transfer_df is not None:
                    positions_margin_change_df = pd.concat([positions_margin_change_df, add_transfer_df])
        if positions_margin_change_df is not None:
            return positions_margin_change_df[['date','asset','amount']]
        return positions_margin_change_df

    def get_futures_position_margin_change_history(self, asset, start_time, end_time):

        startTime = int(start_time.timestamp() * 1000)
        endTime = int(end_time.timestamp() * 1000)

        BASE_URL = 'https://fapi.binance.com'
        PATH = '/fapi/v1/positionMargin/history'


        headers = {
            'X-MBX-APIKEY': self.apiKey
        }
        timestamp = int(time.time() * 1000)

        params = {
            'symbol': asset,
            'recvWindow': 60000,
            'timestamp': timestamp,
            'startTime': startTime,
            'endTime': endTime,
##            'limit': 500,
##           'type':2
        }
        query_string = urllib.parse.urlencode(params)
        params['signature'] = hmac.new(self.secret.encode('utf-8'), query_string.encode('utf-8'),
                                       hashlib.sha256).hexdigest()
        url = urljoin(BASE_URL, PATH)
        print(url)
        r = requests.get(url, headers=headers, params=params)
        dataSet = r.json()
        if dataSet['total']>0:
            transfers = dataSet['rows']

            transfer_df = pd.DataFrame(transfers)

            transfer_df['date'] = pd.to_datetime(transfer_df['timestamp'], unit='ms')

            return transfer_df
        else :
            return None


    def get_future_position_risk(self):


        BASE_URL = 'https://fapi.binance.com'
        PATH = '/fapi/v2/positionRisk'


        headers = {
            'X-MBX-APIKEY': self.apiKey
        }
        timestamp = int(time.time() * 1000)

        params = {
            'recvWindow': 60000,
            'timestamp': timestamp,
        }
        query_string = urllib.parse.urlencode(params)
        params['signature'] = hmac.new(self.secret.encode('utf-8'), query_string.encode('utf-8'),
                                       hashlib.sha256).hexdigest()
        url = urljoin(BASE_URL, PATH)
        print(url)
        r = requests.get(url, headers=headers, params=params)
        dataSet = r.json()
        return dataSet

    def get_futures_transfer_history(self, assets, start_time, end_time):
        transfer_df = None
        for asset in assets:
            if transfer_df is None:
                transfer_df = self.get_coin_futures_transfer_history(asset, start_time, end_time)
            else:
                add_transfer_df = self.get_coin_futures_transfer_history(asset, start_time, end_time)
                if add_transfer_df is not None:
                    transfer_df = pd.concat([transfer_df, add_transfer_df])
        if transfer_df is not None:
            return transfer_df[['date','asset','amount']]
        return transfer_df

    def get_coin_futures_transfer_history(self, asset, start_time, end_time):

        startTime = int(start_time.timestamp() * 1000)
        endTime = int(end_time.timestamp() * 1000)

        BASE_URL = 'https://api.binance.com'
        PATH = '/sapi/v1/futures/transfer'


        headers = {
            'X-MBX-APIKEY': self.apiKey
        }
        timestamp = int(time.time() * 1000)

        params = {
            'asset': asset,
            'recvWindow': 60000,
            'timestamp': timestamp,
            'startTime': startTime,
            'endTime': endTime,
            'limit':30

        }
        query_string = urllib.parse.urlencode(params)
        params['signature'] = hmac.new(self.secret.encode('utf-8'), query_string.encode('utf-8'),
                                       hashlib.sha256).hexdigest()
        url = urljoin(BASE_URL, PATH)
        print(url)
        r = requests.get(url, headers=headers, params=params)
        dataSet = r.json()
        if dataSet['total']>0:
            transfers = dataSet['rows']

            transfer_df = pd.DataFrame(transfers)

            transfer_df['date'] = pd.to_datetime(transfer_df['timestamp'], unit='ms')

            return transfer_df
        else :
            return None

    def get_balance_low_level(self, request_date):
        BASE_URL = 'https://api.binance.com'
        PATH = '/api/v3/account'
        if request_date is None:
            timestamp = int(time.time() * 1000)
        else:
            timestamp = calendar.timegm(request_date.timetuple())

        headers = {
            'X-MBX-APIKEY': self.apiKey
        }
        params = {
            'recvWindow': 5000,
            'timestamp': timestamp,
        }
        query_string = urllib.parse.urlencode(params)
        params['signature'] = hmac.new(self.secret.encode('utf-8'), query_string.encode('utf-8'),
                                       hashlib.sha256).hexdigest()
        url = urljoin(BASE_URL, PATH)
        print(url)
        r = requests.get(url, headers=headers, params=params)
        dataSet = r.json()
        print(dataSet)

    def get_balance(self):
        # Recupérer la balance
        balance = self.exchange.fetchTotalBalance(params={})
        ticker_data = self.exchange.fetchTickers()
        usdt_balance = balance['USDT']
        ETH = float(ticker_data['ETH/USDT']['info']['lastPrice'])
        BTC = float(ticker_data['BTC/USDT']['info']['lastPrice'])
        LTC = float(ticker_data['LTC/USDT']['info']['lastPrice'])
        TRX = float(ticker_data['TRX/USDT']['info']['lastPrice'])
        XRP = float(ticker_data['XRP/USDT']['info']['lastPrice'])
        EOS = float(ticker_data['EOS/USDT']['info']['lastPrice'])
        BCH = float(ticker_data['BCH/USDT']['info']['lastPrice'])
        minBTC = 20 / BTC
        minETH = 20 / ETH
        minLTC = 20 / LTC
        minXRP = 20 / XRP
        minEOS = 20 / EOS
        minTRX = 20 / TRX
        minBCH = 20 / BCH
        return usdt_balance, balance, ticker_data, minBTC, minETH, minLTC, minXRP, minEOS, minTRX, minBCH

    def make_sure_positions_changed(self, positions, sleeping_time = 45):
        position_changed = False
        counter = 1
        while not  position_changed:
            if counter>1:
                print(f'position not reached : reupdating positions {positions}')
            self.cancel_all_orders()
            position_changed = self.change_position(positions)
            # Prints the current time with a five second difference
            print(f'waiting for the oder to be filled {sleeping_time}')
            time.sleep(sleeping_time)
            counter = counter+1

    def update_position(self, pair, pair_signal):
        print(f'entering the global update position for {pair} and signals {pair_signal}')
        print('canceling all orders')
        positions ={}
        if pair == 'BTCUSDT':
            positions['BTC']=pair_signal
            self.make_sure_positions_changed(positions)
        else :
            print(f'pair not handled for the moment {pair}')

    def change_position(self, positions):
        print(f'incoming positions to hodl {positions}')
        usdt_balance, balance, ticker_data, minBTC, minETH, minLTC, minXRP, minEOS, minTRX, minBCH = self.get_balance()
        ETH = float(ticker_data['ETH/USDT']['info']['lastPrice'])
        BTC = float(ticker_data['BTC/USDT']['info']['lastPrice'])
        LTC = float(ticker_data['LTC/USDT']['info']['lastPrice'])
        TRX = float(ticker_data['TRX/USDT']['info']['lastPrice'])
        XRP = float(ticker_data['XRP/USDT']['info']['lastPrice'])
        EOS = float(ticker_data['EOS/USDT']['info']['lastPrice'])
        BCH = float(ticker_data['BCH/USDT']['info']['lastPrice'])
        # calculer avec le last de binance!
        Tbtc = round(positions.get('BTC',0) * balance['USDT'] / BTC, 4)
        Teth = round(positions.get('ETH',0) * balance['USDT'] / ETH, 3)
        Teos = round(positions.get('EOS',0) * balance['USDT'] / EOS, 3)
        Txrp = round(positions.get('XRP',0) * balance['USDT'] / XRP, 0)
        Tltc = round(positions.get('LTC',0) * balance['USDT'] / LTC, 2)
        Ttrx = round(positions.get('TRX',0) * balance['USDT'] / TRX, 0)
        Tbch = round(positions.get('BCH',0) * balance['USDT'] / BCH, 3)
        balance2 = self.exchange.fapiPrivate_get_positionrisk()
        PBTC, PETH, PLTC, PXRP, PTRX, PBCH, PEOS = 0, 0, 0, 0, 0, 0, 0
        for i in range(len(balance2)):
            if balance2[i]['symbol'] == 'BTCUSDT':
                PBTC = balance2[i]['positionAmt']
            elif balance2[i]['symbol'] == 'ETHUSDT':
                PETH = balance2[i]['positionAmt']
            elif balance2[i]['symbol'] == 'LTCUSDT':
                PLTC = balance2[i]['positionAmt']
            elif balance2[i]['symbol'] == 'XRPUSDT':
                PXRP = balance2[i]['positionAmt']
            elif balance2[i]['symbol'] == 'TRXUSDT':
                PTRX = balance2[i]['positionAmt']
            elif balance2[i]['symbol'] == 'EOSUSDT':
                PEOS = balance2[i]['positionAmt']
            elif balance2[i]['symbol'] == 'BCHUSDT':
                PBCH = balance2[i]['positionAmt']
        # trades à faire:
        Ebtc = - float(PBTC) if Tbtc == 0 else round(float(Tbtc) - float(PBTC), 4)
        Eeth = - float(PETH) if Teth == 0 else round(float(Teth) - float(PETH), 2)
        Eeos = - float(PEOS) if Teos == 0 else round(float(Teos) - float(PEOS), 2)
        Eltc = - float(PLTC) if Tltc == 0 else round(float(Tltc) - float(PLTC), 2)
        Exrp = - float(PXRP) if Txrp == 0 else round(float(Txrp) - float(PXRP), 0)
        Etrx = - float(PTRX) if Ttrx == 0 else round(float(Ttrx) - float(PTRX), 0)
        Ebch = - float(PBCH) if Tbch == 0 else round(float(Tbch) - float(PBCH), 0)

        position_threshold_reached = True
        if Ebtc > minBTC:
            print('current BTC position '+str(PBTC))
            print('target BTC position ' + str(Tbtc))
            coursbtc = float(ticker_data['BTC/USDT']['info']['bidPrice']) - 1
            print('creating order buy at '+str(coursbtc) + ' for amount '+str(Ebtc))
            self.exchange.create_order(symbol='BTC/USDT', type='limit', side='buy', amount=Ebtc, price=coursbtc)
            position_threshold_reached = False
        elif Ebtc < -minBTC:
            print('current BTC position '+str(PBTC))
            print('target BTC position ' + str(Tbtc))
            coursbtc = float(ticker_data['BTC/USDT']['info']['askPrice']) + 1
            print('creating order buy at '+str(coursbtc) + ' for amount '+str(Ebtc))
            self.exchange.create_order(symbol='BTC/USDT', type='limit', side='sell', amount=abs(Ebtc), price=coursbtc)
            position_threshold_reached = False

        if Eeth > minETH:
            courseth = float(ticker_data['ETH/USDT']['info']['bidPrice']) - 0.05
            self.exchange.create_order(symbol='ETH/USDT', type='limit', side='buy', amount=Eeth, price=courseth)
            position_threshold_reached = False
        elif Eeth < -minETH:
            courseth = float(ticker_data['ETH/USDT']['info']['askPrice']) + 0.05
            self.exchange.create_order(symbol='ETH/USDT', type='limit', side='sell', amount=abs(Eeth), price=courseth)
            position_threshold_reached = False

        if Eltc > minLTC:
            coursltc = float(ticker_data['LTC/USDT']['info']['bidPrice']) - 0.02
            self.exchange.create_order(symbol='LTC/USDT', type='limit', side='buy', amount=Eltc, price=coursltc)
            position_threshold_reached = False
        elif Eltc < -minLTC:
            coursltc = float(ticker_data['LTC/USDT']['info']['askPrice']) + 0.02
            self.exchange.create_order(symbol='LTC/USDT', type='limit', side='sell', amount=abs(Eltc), price=coursltc)
            position_threshold_reached = False

        if Exrp > minXRP:
            coursxrp = float(ticker_data['XRP/USDT']['info']['bidPrice']) - 0.0001
            self.exchange.create_order(symbol='XRP/USDT', type='limit', side='buy', amount=Exrp, price=coursxrp)
            position_threshold_reached = False
        elif Exrp < -minXRP:
            coursxrp = float(ticker_data['XRP/USDT']['info']['askPrice']) + 0.0001
            self.exchange.create_order(symbol='XRP/USDT', type='limit', side='sell', amount=abs(Exrp), price=coursxrp)
            position_threshold_reached = False

        if Eeos > minEOS:
            courseos = float(ticker_data['EOS/USDT']['info']['bidPrice'])
            self.exchange.create_order(symbol='EOS/USDT', type='limit', side='buy', amount=Eeos, price=courseos)
            position_threshold_reached = False
        elif Eeos < -minEOS:
            courseos = float(ticker_data['EOS/USDT']['info']['askPrice'])
            self.exchange.create_order(symbol='EOS/USDT', type='limit', side='sell', amount=abs(Eeos), price=courseos)
            position_threshold_reached = False

        if Etrx > minTRX:
            courstrx = float(ticker_data['TRX/USDT']['info']['bidPrice'])
            self.exchange.create_order(symbol='TRX/USDT', type='limit', side='buy', amount=Etrx, price=courstrx)
            position_threshold_reached = False
        elif Etrx < -minTRX:
            courstrx = float(ticker_data['TRX/USDT']['info']['askPrice'])
            self.exchange.create_order(symbol='TRX/USDT', type='limit', side='sell', amount=abs(Etrx), price=courstrx)
            position_threshold_reached = False

        if Ebch > minBCH:
            coursbch = float(ticker_data['BCH/USDT']['info']['bidPrice'])
            self.exchange.create_order(symbol='BCH/USDT', type='limit', side='buy', amount=Ebch, price=coursbch)
            position_threshold_reached = False
        elif Ebch < -minBCH:
            coursbch = float(ticker_data['BCH/USDT']['info']['askPrice'])
            self.exchange.create_order(symbol='BCH/USDT', type='limit', side='sell', amount=abs(Ebch), price=coursbch)
            position_threshold_reached = False
        return position_threshold_reached





#!/usr/bin/env python
# coding: utf-8

import pandas as pd
import numpy as np
import statsmodels.api as sm
from statsmodels.api import OLS, add_constant
from statsmodels.tsa.stattools import adfuller
import statsmodels.tsa.stattools as ts
from napoleontoolbox.signal import signal_utility



#                     HRP developed by Jeroen Bouma (https://github.com/JerBouma)                  #
# ================================================================================================ #
#                                                                                                  #
def estimate_long_run_short_run_relationships(y, x):
    """Estimates long-run and short-run cointegration relationship for series y and x.

    Uses a 2-step process to first estimate coefficients for the long-run relationship
        y_t = c + gamma * x_t + z_t

    and then the short-term relationship,
        y_t - y_(t-1) = alpha * z_(t-1) + epsilon_t,

    with z the found residuals of the first equation.

    Parameters
    ----------
    y : pd.Series
        The first time series of the pair to analyse.

    x : pd.Series
        The second time series of the pair to analyse.

    Returns
    -------
    c : float
        The constant term in the long-run relationship y_t = c + gamma * x_t + z_t. This
        describes the static shift of y with respect to gamma * x.

    gamma : float
        The gamma term in the long-run relationship y_t = c + gamma * x_t + z_t. This
        describes the ratio between the const-shifted y and x.

    alpha : float
        The alpha term in the short-run relationship y_t - y_(t-1) = alpha * z_(t-1) + epsilon. This
        gives an indication of the strength of the error correction toward the long-run mean.

    z : pd.Series
        Series of residuals z_t from the long-run relationship y_t = c + gamma * x_t + z_t, representing
        the value of the error correction term.

    """

    assert isinstance(y, pd.Series), 'Input series y should be of type pd.Series'
    assert isinstance(x, pd.Series), 'Input series x should be of type pd.Series'
    assert sum(y.isnull()) == 0, 'Input series y has nan-values. Unhandled case.'
    assert sum(x.isnull()) == 0, 'Input series x has nan-values. Unhandled case.'
    assert y.index.equals(x.index), 'The two input series y and x do not have the same index.'

    long_run_ols = OLS(y, add_constant(x), has_const=True)
    long_run_ols_fit = long_run_ols.fit()

    c, gamma = long_run_ols_fit.params
    z = long_run_ols_fit.resid

    short_run_ols = OLS(y.diff().iloc[1:], (z.shift().iloc[1:]))
    short_run_ols_fit = short_run_ols.fit()

    alpha = short_run_ols_fit.params[0]

    return c, gamma, alpha, z


def engle_granger_two_step_cointegration_test(y, x):
    """Applies the two-step Engle & Granger test for cointegration.

    First fits the long-run relationship
        y_t = c + gamma * x_t + z_t

    and then tests, by Dickey-Fuller phi=1 vs phi < 1 in
        z_t = phi * z_(t-1) + eta_t

    If this implies phi < 1, the z series is stationary is concluded to be
    stationary, and thus the series y and x are concluded to be cointegrated.
    Parameters
    ----------
    y : pd.Series
        the first time series of the pair to analyse

    x : pd.Series
        the second time series of the pair to analyse

    Returns
    -------
    dfstat : float
        The Dickey Fuller test-statistic for phi = 1 vs phi < 1 in the second equation. A more
        negative value implies the existence of stronger cointegration.

    pvalue : float
        The p-value corresponding to the Dickey Fuller test-statistic. A lower value implies
        stronger rejection of no-cointegration, thus stronger evidence of cointegration.

    """

    assert isinstance(y, pd.Series), 'Input series y should be of type pd.Series'
    assert isinstance(x, pd.Series), 'Input series x should be of type pd.Series'
    assert sum(y.isnull()) == 0, 'Input series y has nan-values. Unhandled case.'
    assert sum(x.isnull()) == 0, 'Input series x has nan-values. Unhandled case.'
    assert y.index.equals(x.index), 'The two input series y and x do not have the same index.'

    c, gamma, alpha, z = estimate_long_run_short_run_relationships(y, x)

    # NOTE: The p-value returned by the adfuller function assumes we do not estimate z first, but test
    # stationarity of an unestimated series directly. This assumption should have limited effect for high N,
    # so for the purposes of this course this p-value can be used for the EG-test. Critical values taking
    # this into account more accurately are provided in e.g. McKinnon (1990) and Engle & Yoo (1987).

    adfstat, pvalue, usedlag, nobs, crit_values, icbest = adfuller(z)

    return adfstat, pvalue, crit_values, alpha, z
#                                                                                                  #
# ================================================================================================ #


def apply_cointegration_test(data_df, risk = 10, alpha = 0.05, suffixes=('_btc', '_eth')):

    y,x = data_df['log_return'+suffixes[1]],data_df['log_return'+suffixes[0]]
    adfstat, pvalue, crit_values, alpha_ols, z = engle_granger_two_step_cointegration_test(y,x)

    if pvalue > alpha:
        return False, alpha_ols, z
    elif adfstat > crit_values[str(risk)+'%']:
        return False, alpha_ols, z
    else:
        return True, alpha_ols, z

def merge_series(ts_1, ts_2, suffixes=('_btc', '_eth')):
    pair_df = pd.merge(ts_1, ts_2, how='inner', left_index=True, right_index=True, sort=True,
                             suffixes=suffixes, copy=True, indicator=False, validate=None)
    return pair_df

def add_returns(pair_df, suffixes=('_btc', '_eth')):
    returns_df = pair_df.copy()
    returns_df['return' + suffixes[0]] = returns_df['close' + suffixes[0]].pct_change()
    returns_df['return' + suffixes[1]] = returns_df['close' + suffixes[1]].pct_change()
    returns_df = returns_df.fillna(0)
    return returns_df

def add_log_returns(pair_df, suffixes=('_btc', '_eth')):
    log_returns_df = pair_df.copy()
    log_returns_df['shifted_close'+suffixes[0]] = log_returns_df['close'+suffixes[0]].shift(1)
    log_returns_df['shifted_close'+suffixes[1]] = log_returns_df['close'+suffixes[1]].shift(1)
    # log_returns_df = log_returns_df.dropna()
    log_returns_df = log_returns_df.fillna(method='ffill')
    log_returns_df = log_returns_df.fillna(method='bfill')
    log_returns_df['log_return'+suffixes[0]] = np.log(log_returns_df['close'+suffixes[0]]) - np.log(log_returns_df['shifted_close'+suffixes[0]])
    log_returns_df['log_return'+suffixes[1]] = np.log(log_returns_df['close'+suffixes[1]]) - np.log(log_returns_df['shifted_close'+suffixes[1]])
    return log_returns_df

def compute_log_return_regressor(log_returns_df, log_regressor_kind='log_x_over_log_y', suffixes=('_btc', '_eth')):
    log_return_regressor_df = log_returns_df.copy()

    if log_regressor_kind == 'log_x_over_log_y':
        log_return_regressor_df[log_regressor_kind] = log_return_regressor_df['log_return' + suffixes[0]] / \
                                                      log_return_regressor_df['log_return' + suffixes[1]]
    else:
        log_return_regressor_df[log_regressor_kind] = log_return_regressor_df['log_return' + suffixes[1]] / \
                                                      log_return_regressor_df['log_return' + suffixes[0]]
    log_return_regressor_df = log_return_regressor_df.replace([np.inf, -np.inf], np.nan)
    log_return_regressor_df = log_return_regressor_df.fillna(method='ffill')
    return log_return_regressor_df

def log_return_regressor_half_mean_and_z_score(data_df, log_regressor_kind='log_x_over_log_y', skipping_point=5):
    log_mean_zscore_df = data_df.copy()
    log_mean_zscore_df['log_return_regressor_half_mean'] = log_mean_zscore_df[log_regressor_kind].rolling(skipping_point).mean() / 2
    log_mean_zscore_df['log_return_regressor_zscore'] = signal_utility.expanding_zscore(log_mean_zscore_df[log_regressor_kind], skipping_point=skipping_point,
                                                                                        signal_continuum_threshold=1)
    log_mean_zscore_df = log_mean_zscore_df.fillna(0)
    return log_mean_zscore_df

def compute_zscore_threshold(data_df, threshold_type='max'):
    # cur_df = data_df.copy()
    cur_df = data_df[:-1].copy()
    if threshold_type == 'max':
        row = cur_df.loc[abs(cur_df.log_return_regressor_zscore) == max(abs(cur_df.log_return_regressor_zscore))]
        zscore_threshold = row.log_return_regressor_zscore[0]
    if threshold_type == 'mean':
        zscore_threshold = cur_df.log_return_regressor_zscore.mean()
    return zscore_threshold

def add_ecm_result(data_df, alpha, residuals):
    data_df['ecm'] = alpha*residuals
    return data_df




from napoleontoolbox.mixing import mixing_utility

import lightgbm as lgb

import numpy as np

from sklearn.model_selection import cross_val_score, GridSearchCV, KFold, RandomizedSearchCV, train_test_split



class LGBMModel():

    def __init__(self, preprocessing = False, threshold = 0.6, number_per_year=252):
        self.model = lgb.LGBMRegressor()

    def calibrate(self, X, y):#, method = 'standard'):
        param_grid = {
            'boosting_type': ['gbdt'],
            'num_leaves': [int(x) for x in np.linspace(start=10, stop=30, num=10)],
            'max_depth': [int(x) for x in np.linspace(start=-1, stop=7, num=3)],
            'learning_rate': [0.001,0.01,0.2],
            'n_estimators': [int(x) for x in np.linspace(start=40, stop=120, num=40)]
        }
        gbm_grid = GridSearchCV(self.model, param_grid, cv=2)
        gbm_grid.fit(X.fillna(0), y.fillna(0))
        print('Best parameters found by grid search are:', gbm_grid.best_params_)
        best_params = gbm_grid.best_params_
        self.model.set_params(**best_params)
        return best_params

    def fit(self, X_train, y_train, X_val, y_val):#, method = 'standard'):
        self.model.fit(X_train, y_train, eval_set=[(X_val, y_val)], eval_metric=['l1', 'l2'], early_stopping_rounds=5,verbose=0)

    def predict(self, X_test):#, method = 'standard'):
        y_pred = self.model.predict(X_test)
        return y_pred

    def get_features_importance(self, features_names):#, method = 'standard'):
        run_importances = {}
        for (name, imp) in zip(features_names, self.model.feature_importances_):
            run_importances[name] = imp
        return run_importances



__author__ = "hugo.inzirillo"

from napoleontoolbox.napoleon_config_tools.object.profile import Profile

from napoleontoolbox.napoleon_config_tools.object.environment import Env

__all__ = ["Profile", "Env"]



from napoleontoolbox.rebalancing import allocation
from napoleontoolbox.signal import signal_utility
from napoleontoolbox.utility import metrics
import pandas as pd
import numpy as np


class ERC_Allocation_Model():
    def __init__(self):
        self.w_optim = None
        self.cutting_rate_threshold = 0.7
        self.share_for_flat_signals = 0.1
        return

    def calibrate(self, X, y, number_per_year=252):
        X = X[[col for col in X.columns if 'signal' in col]]
        param_grid = {
            'cutting_rate_threshold': [0.62,0.7,0.85,0.91],
            'share_for_flat_signals': [0.05,0.1,0.22,0.3]
        }
        best_params = {}
        for cutting_rate_threshold in param_grid['cutting_rate_threshold']:
            for share_for_flat_signals in param_grid['share_for_flat_signals']:
                self.fit(X,y,X,y)
                y_pred_discrete = self.predict(X)
                perf_df = signal_utility.reconstitute_prediction_perf(y_pred=y_pred_discrete, y_true=y,
                                                                      transaction_cost=False,
                                                                      print_turnover=False)
                sharpe_strat = metrics.sharpe(perf_df['perf_return'].dropna(), period=number_per_year, from_ret=True)
                best_params[(cutting_rate_threshold,share_for_flat_signals)] = sharpe_strat
        self.cutting_rate_threshold, self.share_for_flat_signals = max(best_params)

    def fit(self, X_train, y_train, X_val, y_val):
        X_t = X_train.copy()
        X_t = X_t.fillna(0)
        n_step, n_asset = X_t.shape
        def filtering_process(series):
            # True if less than 50% of obs. are constant
            return series.value_counts(dropna=False).max() < self.cutting_rate_threshold * n_step
        assets = X_t.apply(filtering_process).values
        w_optim = np.zeros(n_asset)
        signals_returns_df = signal_utility.recompute_perf_returns(X_t.loc[:,assets],y_train)
        signals_returns_df = signals_returns_df.fillna(0)
        w_optim_assets = allocation.ERC(signals_returns_df)
        w_optim_assets = w_optim_assets.reshape(w_optim_assets.size)
        nb_asset_optimized = np.sum(assets)
        nb_asset_not_optimized = n_asset - nb_asset_optimized
        w_optim[assets] = w_optim_assets*(1-self.share_for_flat_signals)
        w_optim[list(map(bool, 1-assets))] = self.share_for_flat_signals * (1 / nb_asset_not_optimized)
        self.w_optim = w_optim

    def predict(self, X_test):
        data = pd.DataFrame(X_test.values * self.w_optim, columns=X_test.columns, index=X_test.index)
        X_test['signal'] = data.sum(axis=1)
        return X_test['signal'].values

    def get_features_importance(self, features_names):
        run_importances = {}
        for (name, imp) in zip(features_names, self.w_optim):
            run_importances[name] = imp
        return run_importances
